//###########################################################
// File: fat.c
//
// For FAT12, FAT16
// Only for first Partition
// Only for drives with 512 bytes per sector (the most)
//
// Based on a White Paper from MS
// FAT: General Overview of On-Disk Format
// Version 1.03, December 6, 2000
//
// MBR MasterBootRecord
// PC intern 4
// M.Tischer
// Data Becker
//
// Organisation of a FAT formatted drive:
// ======================================
// MBR including Partitiontable's
// Hidden Sectors
//
// First Partition
// Bootsector 1
// Reserved Sectors
// FAT1
// FAT2
// ...
// FATn
// RootDirectory (not used if FAT32)
// Data Region
//
// Second Partition
// Bootsector 2
// Reserved Sectors
// .....
//
// Benutzung auf eigene Gefahr !
//
// Use at your own risk !
//
//#########################################################################
// Last change: 29.07.2005
//#########################################################################
// hk@holger-klabunde.de
// http://www.holger-klabunde.de/index.html
//#########################################################################
// Compiler: AVR-GCC 3.4.3
//#########################################################################
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "dos.h"

unsigned char dirbuf[32];   //buffer for directory entrys

unsigned long FirstDataSector; 
unsigned long FirstRootSector; 
//unsigned short RootDirSectors;
unsigned char RootDirSectors=0;   // never saw more then 32 sectors
unsigned char secPerCluster;

U16 FileFirstCluster;     //first cluster of file
unsigned long FileSize;

unsigned long maxsect;           // number of sectors on drive

//###########################################################
//first datacluster is cluster 2 !
unsigned long GetFirstSectorOfCluster(U16 n)
//###########################################################
{
 return (((unsigned long)(n - 2) * secPerCluster) + FirstDataSector);
}

//###########################################################
unsigned char GetDriveInformation(void)
//###########################################################
{
 unsigned char by;
 unsigned long DataSec,TotSec;
 unsigned long bootSecOffset;
 unsigned long FATSz; // FATSize
 unsigned long FATFirstSector;  
 unsigned long CountofClusters;
 U16  i,RootEntrys;

// struct MBR *mbr;
 struct BootSec *boot;
  
 by=IdentifyMedia(); //LaufwerksInformationen holen
 if(by==0)
  {
/*
   by=ReadSector(0,dirbuf); //Lese den MBR. Erster Sektor auf der Platte
                       //enthlt max. 4 Partitionstabellen mit jeweils 16Bytes
                       //Die erste fngt bei 0x01BE an

   mbr=(struct MBR *)dirbuf;
   bootSecOffset=mbr->part1.bootoffset; //Nur den brauche ich
*/

#ifdef COMPACTFLASH_CARD
   CFStartReadSector(0); //Lese den MBR. Erster Sektor auf der Platte
                       //enthlt max. 4 Partitionstabellen mit jeweils 16Bytes
                       //Die erste fngt bei 0x01BE an
   CF_CS_OFF();

   bootSecOffset=0; //Nur den brauche ich
   for(i=0; i<BYTE_PER_SEC; i++) //Sektor mu aber komplett abgeholt werden
    {
     CF_RD_OFF();
     NOP; //Schon bei 8MHz mu man hier etwas warten
     NOP; //Es knnte evtl. ntig sein hier noch mehr
        //nop einzufgen, um mglichst viele CF zu lesen
        //Bei mir reichten bisher 2 nop
        //Sicherheitshalber noch einen mehr
     NOP;

#ifdef F_16 //16MHz processor clock
     NOP;
#endif
   
     if(i==0x1C6) { bootSecOffset+=CF_READ_DATA(); }
     if(i==0x1C7) { bootSecOffset+=((unsigned long)CF_READ_DATA())<<8; }
     if(i==0x1C8) { bootSecOffset+=((unsigned long)CF_READ_DATA())<<16; }
     if(i==0x1C9) { bootSecOffset+=((unsigned long)CF_READ_DATA())<<24; }
     CF_RD_ON();
    }

   CF_CS_ON();

//IMPORTANT NOTICE:
//Set bootSecOffset to zero here if your card does not have a partitiontable !

   CFStartReadSector(bootSecOffset); //Lese den Bootsektor
   CF_CS_OFF();

   for(i=0; i<BYTE_PER_SEC; i++)
    {
     CF_RD_OFF();
     NOP; //Schon bei 8MHz mu man hier etwas warten
     NOP; //Es knnte evtl. ntig sein hier noch mehr
        //nop einzufgen, um mglichst viele CF zu lesen
        //Bei mir reichten bisher 2 nop
        //Sicherheitshalber noch einen mehr
     NOP;

#ifdef F_16 //16MHz processor clock
     NOP;
#endif
   
     if(i>12 && i<36) dirbuf[i-13]=CF_READ_DATA();
     CF_RD_ON();
    }

   CF_CS_ON();
#endif

#ifdef MMC_CARD_SPI
   MMC_CS_OFF();
   MMCStartReadSector(0); //Lese den MBR. Erster Sektor auf der Platte
                       //enthlt max. 4 Partitionstabellen mit jeweils 16Bytes
                       //Die erste fngt bei 0x01BE an, und nur die nehme ich !
   
   bootSecOffset=0; //Nur den brauche ich
   for(i=0; i<BYTE_PER_SEC; i++) //Sektor mu aber komplett abgeholt werden
    {
     SPI_WRITE(0xFF);
     if(i==0x1C6) { bootSecOffset+=SPDR; }			  //
     if(i==0x1C7) { bootSecOffset+=((unsigned long)SPDR)<<8; } //
     if(i==0x1C8) { bootSecOffset+=((unsigned long)SPDR)<<16; } //
     if(i==0x1C9) { bootSecOffset+=((unsigned long)SPDR)<<24; } //
    }

   SPI_WRITE(0xFF); //Response abholen
   SPI_WRITE(0xFF); //Response abholen

//IMPORTANT NOTICE:
//Set bootSecOffset to zero here if your card does not have a partitiontable !
   MMCStartReadSector(bootSecOffset); //Lese den Bootsektor

   for(i=0; i<BYTE_PER_SEC; i++)
    {
     SPI_WRITE(0xFF);
     if(i>12 && i<36) dirbuf[i-13]=SPDR;
    }

   SPI_WRITE(0xFF); //CRC
   SPI_WRITE(0xFF);
   MMC_CS_ON();
#endif

   boot=(struct BootSec *)dirbuf;
   
   secPerCluster=boot->BPB_SecPerClus; //Sectors per Cluster
   RootEntrys=boot->BPB_RootEntCnt;    //32 Byte Root Directory Entrys

//Number of sectors for FAT
//   if(boot->BPB_FATSz16 != 0) FATSz = boot->BPB_FATSz16;
//   else FATSz = boot->eb.rm32.BPB_FATSz32; //Fr FAT32

   FATSz = boot->BPB_FATSz16;

   RootDirSectors = (unsigned char)( ((RootEntrys * 32) + (BYTE_PER_SEC - 1)) / BYTE_PER_SEC);
   
   FATFirstSector= bootSecOffset + boot->BPB_RsvdSecCnt;
   FirstRootSector = FATFirstSector + (boot->BPB_NumFATs * FATSz);
   FirstDataSector = FirstRootSector + RootDirSectors;

   if(boot->BPB_TotSec16 != 0) TotSec = boot->BPB_TotSec16;
   else TotSec = boot->BPB_TotSec32;

//Number of data sectors
   DataSec = TotSec - (boot->BPB_RsvdSecCnt + (boot->BPB_NumFATs * FATSz) + RootDirSectors);

//Number of valid clusters
   CountofClusters = DataSec / secPerCluster;
//   maxcluster=CountofClusters+2;

//Note also that the CountofClusters value is exactly that: the count of data clusters
//starting at cluster 2. The maximum valid cluster number for the volume is
//CountofClusters + 1, and the "count of clusters including the two reserved clusters"
// is CountofClusters + 2.

//Das ist laut MS der einzige Weg FAT12,16,32 zu erkennen.
//Der File System String im Bootsektor wird NICHT ausgewertet
   if(CountofClusters < 65525)
     {
     }
   else
     {
      return F_ERROR; // Wrong FAT System
     }

  }
 else
  {
   return F_ERROR; // CF gives no answer
  } 

 return F_OK;
}

